#-----------------------------------------------------------
# Script Name: prepare_spygen.R
# Purpose of script:
#   Create all_samples.csv to analyse edna sequencing data
# Author: Morgane Bruno
# Contact: morgane.bruno@cefe.cnrs.fr
# Date Created: 2022-07-20
#
# Usage: Rscript prepare_spygen.R [options]
# 
# Options:
#   -d DAT_FOLDER, --dat_folder=DAT_FOLDER
#     Folder with dat files
#   
#   -c CORR_TAGS.XLSX, --ctf=CORR_TAGS.XLSX
#     SPYGEN file, correspondance between tags and samples
#   
#   -p PROJECT_NAME, --project_name=PROJECT_NAME
#     Name of the project (ex: 'Med')
#   
#   -f FQ_FOLDER, --fq_folder=FQ_FOLDER
#     Folder with .fastq.gz files
#   
#   -o ALL_SAMPLES.CSV, --out_file=ALL_SAMPLES.CSV
#     Output file
#
#   -h, --help
#-----------------------------------------------------------

#_libs
suppressMessages(library(tidyverse))
library(readxl)
library(optparse)
library(stringi)
library(stringr)

#_main
main <- function() {
  # Read arguments
  args <- get_args()
  # List all sheet names in an excel spreadsheet (i.e. marker)
  marker_sheet <- readxl::excel_sheets(args$ctf)
  # Print arguments (Starting message)
  print_args(args, marker_sheet)
  # Read data
  dat <- read_dat_files(args$dat_folder, marker_sheet)
  corr_tags <- read_alice_file(args$ctf, marker_sheet)
  # Check if tags are unique
  check_tags(corr_tags, args$ctf)
  # Create all samples
  all_samples <- create_all_samples(corr_tags, dat, args$project_name, args$fq_folder)
  # Write all_samples.csv
  utils::write.table(x = all_samples, file = args$out_file, col.names = FALSE,
                     row.names = FALSE, sep = ";", quote = FALSE)
}

#_fun
get_args <- function() {
  option_list <- list(
    optparse::make_option(c("-d", "--dat_folder"), type = "character",
                          default = NULL,
                          help = "Folder with dat files"),
    optparse::make_option(c("-c", "--ctf"), type = "character",
                          default = NULL,
                          help = "SPYGEN file, correspondance between tags and samples",
                          metavar = "CORR_TAGS.XLSX"),
    optparse::make_option(c("-p", "--project_name"), type = "character",
                          default = NULL,
                          help = "Name of the project (ex: 'Med')"),
    optparse::make_option(c("-f", "--fq_folder"), type = "character",
                          default = NULL,
                          help = "Folder with .fastq.gz files"),
    optparse::make_option(c("-o", "--out_file"), type = "character",
                          default = NULL,
                          help = "Output file",
                          metavar = "ALL_SAMPLES.CSV")
  )
  opt_parser <- optparse::OptionParser(option_list = option_list)
  opt <- optparse::parse_args(opt_parser)

  # missing option
  if (is.null(opt$dat_folder)) {
    optparse::print_help(opt_parser)
    stop("Arg --dat_folder [-d] is required")
  }
  if (is.null(opt$ctf)) {
    optparse::print_help(opt_parser)
    stop("Arg --ctf [-c] is required")
  }
  if (is.null(opt$project_name)) {
    optparse::print_help(opt_parser)
    stop("Arg --project_name [-p] is required")
  }
  if (is.null(opt$fq_folder)) {
    optparse::print_help(opt_parser)
    stop("Arg --fq_folder [-f] is required")
  }
  if (is.null(opt$out_file)) {
    optparse::print_help(opt_parser)
    stop("Arg --out_file [-o] is required")
  }
  return(opt)
}

create_all_samples <- function(corr_tags, dat, project_name, fq_folder) { #no lint
  # Create all_samples dataframe (used for demultiplexing)
  # Args:
  #   - corr_tags (tibble): correspondance between tags and samples
  #   - dat (tibble): metadata, tags/marker
  #   - project_name (string): name of the project
  # Output:
  #   - tibble: all_samples
  all_samples <- join_metadata(corr_tags, dat) %>% # join dat and corr_tags
    replace_na_sample(.) %>% # NA sample == tag used as blank (not used in the experiment)
    dplyr::mutate(PROJECT = dplyr::case_when(  # add column "PROJECT" (Other, Blank or project_name)
      grepl("Blank", SAMPLE) ~ "Blank",
      grepl("Other", SAMPLE, ignore.case = TRUE) ~ "Other",
      TRUE ~ project_name
    ))
  # Check if samples are unique
  check_samples(all_samples)
  # Rename samples (avoids duplicate when merging several projects)
  all_samples <- all_samples %>% dplyr::mutate(SAMPLE_LONG = dplyr::case_when(
    grepl("SPY", SAMPLE) ~ SAMPLE,
    grepl("Blank", SAMPLE) ~ SAMPLE,
    grepl("Other", SAMPLE, ignore.case = TRUE) ~ paste("Other", RUN, stringr::str_extract(SAMPLE, "[0-9]*_[0-9]*"), sep = "_"),
    TRUE ~ paste(stringr::word(string = SAMPLE, start = 1, end = 1, sep = '_'), RUN,
                 stringr::word(string = SAMPLE, start = 2, end = 2, sep = '_'), sep = '_')
  ))
  # Rename runs: replace short run name by full run name
  all_samples <- rename_runs(all_samples, fq_folder) %>%
    dplyr::select(P1, RUN_LONG, SAMPLE_LONG, PROJECT, MARKER)
}

check_samples <- function(all_samples) {
  if (is_unique(all_samples, "SAMPLE")) {
    message("All samples in all_samples.csv are unique.")
  } else {
    print(duplicated_data(all_samples, "SAMPLE"), n = 25)
    stop("all_samples.csv contains duplicated samples for the same marker (See table above).\n")
  }
}

check_tags <- function(corr_tags, corr_tags_file) {
  if (is_unique(corr_tags, "TAG")) {
    message(paste("All tags in", corr_tags_file, "are unique."))
  } else {
    print(duplicated_data(corr_tags, "TAG"), n = 25)
    stop(paste(corr_tags_file,
               "contains duplicated tags for the same marker and run (See table above).\n"))
  }
}

read_dat_files <- function(dat_folder, marker_sheet) {
  # Read marker.dat files
  # Args:
  #   - dat_folder (string): path to the folder containing .dat files
  #   - marker_sheet (vector): all sheet names (i.e list of marker)
  # Output:
  #   - tibble
  dat <- do.call('rbind',
                 lapply(marker_sheet, function(one_marker) {
                   # check if dat file exist
                   file_name <- list.files(dat_folder, pattern = paste0("(?i)", one_marker,".dat"))
                   if (rlang::is_empty(file_name)) {
                     stop(paste0(one_marker, ": dat file is missing"))
                   }
                   dat_file <- paste0(dat_folder, file_name)
                   # read dat file
                   utils::read.csv(dat_file, header = FALSE, sep = '\t') %>%
                     dplyr::tibble() %>%
                     dplyr::mutate(MARKER = one_marker)
                 })
  )
  colnames(dat) <- c("P1", "P2", "TAG", "PRIMER F", "PRIMER R", "F", "MARKER")
  return(dat)
}

read_alice_file <- function(corr_tags_file, marker_sheet) {
  # Read Corr_tags.csv file
  # Args:
  #   - corr_tags_file (string): Corr_tags.csv
  #   - marker_sheet (vector): all sheet names (i.e list of marker)
  # Output:
  #   - tibble: Corr_tags
  do.call('rbind',
          lapply(marker_sheet, function(one_marker) {
            readxl::read_xlsx(corr_tags_file, sheet = one_marker) %>%
              dplyr::tibble() %>%
              dplyr::mutate(MARKER = one_marker) %>%
              dplyr::select(-TAG_NAME)
          })
  )}

rename_runs <- function(all_samples, fq_folder) {
  fq_files <- list.files(fq_folder, pattern = "_R1.fastq.gz")
  full_name_runs <- sub("_R1.fastq.gz", "", fq_files)
  all_samples$RUN_LONG <- sapply(all_samples$RUN, function(run) {full_name_runs[grep(run, full_name_runs)]})
  check_fastq_runs(all_samples, full_name_runs)
  return(all_samples)
}

check_fastq_runs <- function(all_samples, full_name_runs) {
  runs_unique <- all_samples$RUN %>% unique()
  check_runs <- do.call('rbind', 
                        lapply(runs_unique, function(run) { 
                          tibble(RUN = run, PRESENT = grepl(run, full_name_runs) %>% any()) 
                        })
  )
  if (FALSE %in% check_runs$PRESENT) {
    print(check_runs %>% dplyr::filter(PRESENT == FALSE) %>% dplyr::select(RUN))
    stop("Some of the fastq files are missing (See table above).")
  } else {
    message("All runs in metadata have a corresponding fastq file.")
  }

  check_fastq <- do.call('rbind', 
                         lapply(full_name_runs, function(run) { 
                           tibble(RUN = run, PRESENT = run %in% all_samples$RUN_LONG) 
                         })
  )
  if (FALSE %in% check_runs$PRESENT) {
    warnings("Not all fastq files are present in metadata.")
    print(check_fastq %>% dplyr::filter(PRESENT == FALSE) %>% dplyr::select(RUN))
  } else {
    message("All fastq files are present in metadata.")
  }
}

join_metadata <- function(corr_tags, dat) {
  corr_tags_splitted <- corr_tags %>% group_split(RUN)
  do.call('rbind',
          lapply(corr_tags_splitted, function(one_run){
            dat %>% dplyr::left_join(one_run, by = c("TAG", "PRIMER F", "PRIMER R", "MARKER")) %>%
              dplyr::select(-c("P2", "F")) %>%
              dplyr::mutate(RUN = one_run$RUN %>% unique())
          })
  )
}

replace_na_sample <- function(all_samples) {
  all_samples_splitted <- all_samples %>% dplyr::group_split(MARKER)
  do.call('rbind',
          lapply(all_samples_splitted, function(one_marker) {
            na_nb <- one_marker %>% 
              dplyr::filter(is.na(SAMPLE)) %>% 
              dplyr::group_by(RUN) %>% 
              dplyr::summarise(n = n())
            marker <- one_marker$MARKER %>% unique()
            blank <- apply(X = na_nb, MARGIN = 1, FUN = function(one_row){
              paste("Blank", marker, one_row["RUN"], seq(1:one_row["n"]), sep = "_")
            }) %>% unlist()
            one_marker[is.na(one_marker$SAMPLE),]$SAMPLE <- blank
            return(one_marker)
          })
  )
}

is_unique <- function(data, column_name) { # nolint
  # Check if tags used in one sequencing run for one marker is unique.
  # Or if sample names are unique for one marker
  # Args:
  #   - data (tibble): corr_tags or all_samples
  #   - column_name (string): "TAG" or "SAMPLE"
  # Output:
  #   - boolean: TRUE if tags/samples are unique
  if (column_name == "TAG") {
    group <- c("MARKER", "RUN")
  } else {
    group <- c("MARKER")
  }
  !(data %>% 
      dplyr::group_by_at(group) %>%
      dplyr::select(dplyr::matches(c(group, column_name))) %>%
      duplicated() %>%
      any())
}

duplicated_data <- function(data, column_name) { # nolint
  # Return the list of duplicated tags/samples
  # Args:
  #   - data (tibble): corr_tags or all_samples
  #   - column_name (string): "TAG" or "SAMPLE"
  # Output:
  #   - tibble: duplicated tag
  if (column_name == "TAG") {
    group <- c("MARKER", "RUN")
  } else {
    group <- c("MARKER")
  }
  data %>% 
    dplyr::group_by_at(c(group, column_name)) %>%
    dplyr::summarise(n = n(), .groups = "drop_last") %>%
    dplyr::filter(n > 1)
}

print_args <- function(args, marker_sheet) {
  # Starting message
  message("===================================================")
  message(paste("Project:", args$project_name,
                "\nSpygen file:", args$ctf, "\nMarker:"))
  message(paste("-", marker_sheet, "\n"))
  message(paste("Output file:", args$out_file))
  message("===================================================")
}

if (!interactive()) {
  main()
}
